from typing import Any

from pytest import raises
from triad.utils.dispatcher import (
    conditional_broadcaster,
    conditional_dispatcher,
    run_at_def,
)
from pkg_resources import EntryPoint, get_distribution
import cloudpickle


def test_run_at_def():
    data = []

    @run_at_def
    def f1():
        """test1"""
        data.append(1)

    assert [1] == data
    assert "test1" == f1.__doc__

    @run_at_def()
    def f2():
        data.append(1)

    assert [1, 1] == data

    @run_at_def(x=2)
    def f3(x) -> int:
        """test3"""
        data.append(x)

    assert [1, 1, 2] == data
    assert "test3" == f3.__doc__

    @run_at_def()
    def f4(a=3) -> int:
        """test4"""
        data.append(a)

    assert [1, 1, 2, 3] == data
    assert "test4" == f4.__doc__


def test_conditional_dispatcher():
    @conditional_dispatcher
    def f1(a: Any, b: int) -> int:
        """test1"""
        raise NotImplementedError

    assert "test1" == f1.__doc__

    @conditional_dispatcher(entry_point="invalid")
    def f2(a: Any, b: int) -> int:
        """test2"""
        raise NotImplementedError

    assert "test2" == f2.__doc__

    def _bad_matcher(a, *args, **kwargs):
        raise Exception

    f1.register(
        matcher=lambda a, b, **kwargs: isinstance(a, str) and isinstance(b, int),
        func=lambda a, b: len(a) + b,
    )
    f1.register(
        matcher=_bad_matcher,  # this should be ignored
        func=lambda a, b: len(a) + b,
    )
    f1.register(
        matcher=lambda a, *args, **kwargs: isinstance(a, str)
        and isinstance(args[0], int),
        func=lambda a, b: len(a) + b + 5,
    )

    @f1.candidate(lambda a, b: isinstance(a, int) and isinstance(b, int))
    def f3(a: int, b: int):
        return a + b

    f1.register(
        matcher=lambda a, b, **kwargs: isinstance(a, str) and isinstance(b, int),
        func=lambda a, b: len(a) + b - 1,
        priority=2,
    )

    assert 5 == f1(2, 3)
    # higher priority
    assert 6 == f1("abcd", 3)

    f1.register(
        matcher=lambda a, *args, **kwargs: isinstance(a, str)
        and isinstance(args[0], int),
        func=lambda a, b: len(a) + b - 2,
        priority=2,
    )

    # newer registration
    assert 5 == f1("abcd", 3)

    with raises(NotImplementedError):
        f1([0], True)

    with raises(NotImplementedError):
        f2(True, True)

    @f1.candidate(lambda a, b: isinstance(a, int) and isinstance(b, int))
    def f4(a: int, b: int):
        return a**b

    ff1 = cloudpickle.loads(cloudpickle.dumps(f1))
    assert 8 == ff1(2, 3)

    @f1.candidate(lambda a, b: isinstance(a, int) and isinstance(b, int))
    def f5(a: int, b: int):
        return a * b + 1

    assert 7 == f1(2, 3)

    # ignore case
    @f1.candidate(lambda a, b: isinstance(a, str) and isinstance(b, str), priority=0)
    def f6(a, b):
        return 6

    @f1.candidate(lambda a, b: isinstance(a, str) and isinstance(b, str), priority=0)
    def f7(a, b):  # this should take no effect
        return 7

    assert 6 == f1("xx", "yy")

    f1.register(  # this should overwrite
        matcher=lambda a, *args, **kwargs: isinstance(a, str)
        and isinstance(args[0], int),
        func=lambda a, b: len(a) + b - 10,
        priority=2.0,
    )

    assert -3 == f1("abcd", 3)


def test_conditional_broadcaster():
    @conditional_broadcaster
    def f1(a=1) -> int:
        """test1"""
        raise NotImplementedError

    assert "test1" == f1.__doc__

    @conditional_broadcaster(entry_point="invalid")
    def f2(a: Any) -> int:
        """test2"""
        raise NotImplementedError

    assert "test2" == f2.__doc__

    data = []

    @f1.candidate(lambda x: x < 10, priority=1)
    def f11(a):
        data.append(a + 1)

    @f1.candidate(lambda x: x < 10, priority=2)
    def f12(a):
        data.append(a + 2)

    f1(5)
    assert [7, 6] == data

    with raises(NotImplementedError):
        f1(11)  # no matching

    with raises(NotImplementedError):
        f2(10)  # no registration


def test_preload(mocker):
    mocker.patch(
        "triad.utils.entry_points._entry_points_for",
        return_value=[
            EntryPoint.parse(
                "dummy=tests.utils.dispatcher_examples.examples",
                get_distribution("triad"),
            ),
            EntryPoint.parse(
                "dummy=tests.utils.dispatcher_examples.examples:register2",
                get_distribution("triad"),
            ),
            EntryPoint.parse(
                "dummy2=tests.utils.dispatcher_examples.invalid",
                get_distribution("triad"),
            ),
        ],
    )

    from tests.utils.dispatcher_examples import dtest

    assert 1 == dtest(1)
    assert 2 == dtest("ab")
    assert 2 == dtest(dict(a=1, b=2))
    assert 100 == dtest(None)



from __future__ import absolute_import

from triad.utils.hash import to_uuid


def test_to_uuid():
    assert to_uuid() == to_uuid()
    assert to_uuid(None) == to_uuid(None)
    assert to_uuid() != to_uuid(None)

    d = {"a": 1, "b": {"x": "x", "y": [1, 2, 3]}}
    id1 = to_uuid(d)
    id2 = to_uuid(d)
    d["b"]["y"][0] = None
    id3 = to_uuid(d)
    d["a"] = "1"
    id4 = to_uuid(d)
    assert id1 == id2
    assert id1 != id3
    assert id4 != id3

    id1 = to_uuid([0, 1])
    id2 = to_uuid([1, 0])
    id3 = to_uuid(x for x in [0, 1])
    assert id1 != id2
    assert id1 == id3

    id1 = to_uuid(["a", "aa"])
    id2 = to_uuid(["aa", "a"])
    assert id1 != id2

    assert to_uuid([Mock2(), Mock2()]) == to_uuid([Mock(), Mock()])
    assert to_uuid([Mock2(2), Mock2(3)]) == to_uuid([Mock(2), Mock(3)])
    assert to_uuid([Mock2(2), Mock2(3)]) != to_uuid([Mock(3), Mock(3)])


class Mock(object):
    def __init__(self, n=1):
        self.n = n

    def __uuid__(self) -> str:
        return str(self.n)


class Mock2(object):
    def __init__(self, n=1):
        self.n = n

    def __uuid__(self) -> int:
        return self.n



from typing import Any, Callable, Dict, Iterable, List, Optional
from copy import deepcopy
import pandas as pd
from pytest import raises

from triad.collections.function_wrapper import (
    AnnotatedParam,
    FunctionWrapper,
    NoneParam,
    OtherParam,
    function_wrapper,
)
from triad.exceptions import InvalidOperationError
from triad import to_uuid


class _Dummy:
    pass


@function_wrapper(None)
class MockFunctionWrapper(FunctionWrapper):
    pass


@MockFunctionWrapper.annotated_param(pd.DataFrame, "d", child_can_reuse_code=True)
class DfParam(AnnotatedParam):
    pass


@MockFunctionWrapper.annotated_param(_Dummy)
class Df2Param(DfParam):
    pass


@MockFunctionWrapper.annotated_param("[Series]", "s", lambda a: a == pd.Series)
class SeriesParam(AnnotatedParam):
    pass


def test_registration():
    with raises(InvalidOperationError):

        @function_wrapper(None)
        class _F1:  # not a subclass of AnnotatedParam
            pass

    with raises(InvalidOperationError):

        @FunctionWrapper.annotated_param(_Dummy, ".")
        class _D1:  # not a subclass of AnnotatedParam
            pass

    with raises(InvalidOperationError):

        @FunctionWrapper.annotated_param(
            _Dummy
        )  # _NoneParam doesn't allow child to reuse the code
        class _D2(NoneParam):
            pass


def test_misc():
    f = deepcopy(FunctionWrapper(f2))
    assert f(1, 2, c=3, b=4) == 1 + 2 + 4 - 3

    f = deepcopy(FunctionWrapper(f1))
    assert f.input_code == "xx"
    assert f.output_code == "n"

    f = deepcopy(MockFunctionWrapper(f1))
    assert f.input_code == "ds"
    assert f.output_code == "n"

    f = MockFunctionWrapper(f6)
    assert isinstance(f._params["a"], Df2Param)


def test_uuid():
    x = FunctionWrapper(f2)
    y = deepcopy(x)
    z = FunctionWrapper(f5)
    assert to_uuid(x) == to_uuid(y)
    assert to_uuid(y) != to_uuid(z)


def test_parse_annotation():
    p = FunctionWrapper.parse_annotation(None)
    assert p.code == "x"
    assert isinstance(p, OtherParam)
    p = FunctionWrapper.parse_annotation(type(None))
    assert p.code == "x"
    assert isinstance(p, OtherParam)
    p = FunctionWrapper.parse_annotation(type(None), none_as_other=False)
    assert p.code == "n"
    assert isinstance(p, NoneParam)

    p = MockFunctionWrapper.parse_annotation(pd.DataFrame)
    assert p.code == "d"

    p = MockFunctionWrapper.parse_annotation(pd.Series)
    assert p.code == "s"


def test_parse_function():
    def _parse_function(f, params_re, return_re):
        MockFunctionWrapper(f, params_re, return_re)

    _parse_function(f1, "^ds$", "n")
    raises(TypeError, lambda: _parse_function(f1, "^xx$", "n"))
    _parse_function(f2, "^xxxx$", "n")
    _parse_function(f3, "^yz$", "n")
    _parse_function(f4, "^0x$", "d")
    _parse_function(f6, "^d$", "n")
    _parse_function(f7, "^yz$", "n")


def f1(a: pd.DataFrame, b: pd.Series) -> None:
    pass


def f2(e: int, a, b: int, c):
    return e + a + b - c


def f3(*args, **kwargs):
    pass


def f4(self, a) -> pd.DataFrame:
    pass


def f5(e: int, a, b: int, c):
    return e + a + b - c


def f6(a: _Dummy) -> None:
    pass


def f7(*args: Any, **kwargs: int):
    pass



# flake8: noqa
from triad_version import __version__

from triad.collections import FileSystem, IndexedOrderedDict, ParamDict, Schema
from triad.utils import (
    SerializableRLock,
    assert_arg_not_none,
    assert_or_throw,
    conditional_broadcaster,
    conditional_dispatcher,
    extensible_class,
    extension_method,
    make_empty_aware,
    run_at_def,
    run_once,
    to_uuid,
)
